--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: DEBUG
-- @author    Valentin Schmidt
-- @version   0.4
--
-- Debug library that shows error information and a stack trace when a scripting
-- error occurs, both for imported/included Lingo/LSW and JS scripts.
--
-- It also allows to add breakpoints to your code, which will show the current
-- stack trace. A breakpoint is created by adding the line "--{B}" for
-- Lingo/LSW scripts, and "//{B}" for JS scripts.
--
-- NOTES:
-- This is a special lib that is NOT meant to be imported via $.import()!
-- Instead, its "injected" (as ancestor) into the LSW core object by passing
-- comandline argument "-debug" to lsw.exe.
--
-- To work properly for Lingo/LSW scripts, it's required to use the keyword
-- "return" (for returning from a handler) exclusively in lower-case, and to 
-- use a NOT lower-case version (like RETURN or Return) for the RETURN 
-- constant that stands for numToChar(13).
-- 
--****************************************************************************

property $

-- config
property stackShowMax
property stopOnError

-- private
property _regex
property _file

property _callStack
property _dbgFlag
property _errObj
property _traceFile
property _dbgCode
property _dbgTry
property _dbgCatch

property _breakPointLingo
property _breakPointJS

-- regexp constants
property _rexLingoFunctionStart
property _rexLingoFunctionEnd
property _rexLingoReturn
property _rexJsFunctionStart
property _rexJsFunctionEnd

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
	
  -- private lib
  me._regex = me.importAndForget("regex")
	me._file = me.importAndForget("file")
	
  -- config
  me.stackShowMax = 50
  me.stopOnError = FALSE

  -- properties for providing a debug callstack
  me._callStack = []
  me._dbgFlag = TRUE
  me._errObj = VOID

	me._traceFile = $.PATH&"~trace.log"
	
  -- template for code added to beginning of each Lingo handler
  me._dbgCode = "c=[:]"&RETURN&\
    "c[#name]="&QUOTE&"{{1}}"&QUOTE&RETURN&\
    "c[#file]="&QUOTE&"{{2}}"&QUOTE&RETURN&\
    "c[#handler]="&QUOTE&"{{3}}"&QUOTE&RETURN&\
    "c[#params]=[]"&RETURN&\
    "repeat with i = {{4}} to the paramCount"&RETURN&\
    "  c[#params].add(param(i))"&RETURN&\
    "end repeat"&RETURN&\
    "_global.$.debug.push(c)"

  -- templates of code added to beginning and end of each JS function
  me._dbgTry = "try{"
  me._dbgCatch = "}catch(e){__locals=propList();for(i in arguments.callee)__locals.addProp(i+'', eval(i)+'');_global.$.debug.traceJsError(arguments.callee.name,e,e.stack.split('\n'),scriptfile,__locals);}"

  me._breakPointLingo = "_movie.traceLogFile="&QUOTE&me._traceFile&QUOTE&RETURN&\
  "showLocals()"&RETURN&\
  "_movie.traceLogFile="&QUOTE&QUOTE&RETURN&\
  "_global.$.debug.lingoBreak()"&RETURN

  me._breakPointJS = "__e=new Error('dummy');__locals=propList();for(i in arguments.callee)__locals.addProp(i+'', eval(i)+'');_global.$.debug.jsBreak(arguments.callee.name,__e.stack.split('\n'),scriptfile,__locals,1);"


  -- regular expessions

  -- Lingo/LSW
  me._rexLingoFunctionStart = "^\s*(on[ \t]+(\w*)(.*))$"
  me._rexLingoReturn = "return([ \t\(]+.*)$"
  me._rexLingoFunctionEnd = "^end\s*$"

  -- JS
  me._rexJsFunctionStart = "^(function (\w*)(.*))"
  me._rexJsFunctionEnd = "^(})"

  -- setup alertHook
  _player.alertHook = me

  return me
end

----------------------------------------
-- 
----------------------------------------
on destroy (me)
	me._file.delete(me._traceFile)
	me.ancestor.destroy()
end

----------------------------------------
-- pushes entry to lingo debug trace stack
----------------------------------------
on push (me, c)
  if me._dbgFlag then
    me._callStack.add(c)
  end if
end

----------------------------------------
-- pops entry from lingo debug trace stack
----------------------------------------
on pop (me)
  if me._dbgFlag AND me._callStack.count then
    me._callStack.deleteAt(me._callStack.count)
  end if
end

----------------------------------------
-- pops entry from lingo debug trace stack and returns original return value
----------------------------------------
on r (me, val)
  me.pop()
  return val
end

----------------------------------------
-- Overwrites LSW's import(), adding debugging extensions
----------------------------------------
on import (me, tLibName, tArgs)

  fp = xtra("fileIO").new()

  tLibFolder = "lib"& me.PD & tLibName

  found = FALSE
  repeat with dir in _player.searchPathList
    fn = dir & tLibFolder

    -- is lib already imported?
    if not voidP(me._.libdict[fn]) then
      me[tLibName] = me._.libdict[fn]
      return me._.libdict[fn]
    end if

    -- does folder exist?
    fp.openFile(fn, 1)
    found = (fp.status()=-36)
    fp.closeFile() -- just in case it was a file
    if found then
      tLibFolder = fn
      exit repeat
    end if
  end repeat

  -- look for index file
  repeat with tType in ["ls", "js"]
    tScriptFile = tLibFolder & me.PD & "index." & tType
    fp.openFile(tScriptFile, 1)
    found = (fp.status()=0)
    if found then exit repeat
  end repeat
  if not found then return FALSE -- no index file found!

  -- read and close file
  tCode = fp.readFile()
  fp.closeFile()
  fp = 0

  -- create new script
  m = new(#script, castlib("lswlib"))
  m.scriptType = #parent -- only parent allowed!
  if tType="js" OR tType="jsw" then
    m.scriptSyntax = #javascript
  else -- ls, lsw
    m.scriptSyntax = #lingo
  end if
  m.name = tLibName

  -- DEBUG ###############################################################################

  if m.scriptSyntax=#lingo then

    pattList = ["\{\{1\}\}", "\{\{2\}\}", "\{\{4\}\}"]
    replStrList = [m.name, tScriptFile, "2"]
    tDbgCode = me._regex.replace_multi(pattList, replStrList, me._dbgCode)

    -- save object where error occured, so properties can be listed in alterHook
    put RETURN&"if _global.$.debug._dbgFlag then _global.$.debug._errObj=param(1)" after tDbgCode

    -- add debug code to all hander beginnings
    tCode = me._regex.replace_callback(me._rexLingoFunctionStart, tCode, #replace, me, "gmi", tDbgCode)

		-- replace "return foo" with "return"&" _global.$.debug.r("&foo&")"
  	-- note: i doesn't work for "return" vs "RETURN"
    tCode = me._regex.replace(me._rexLingoReturn, "return"&" _global.$.debug.r($1)", tCode, "gm")

    pattList = []
    replStrList = []

    -- precede function end "end" width "_global.$.debug.pop()"
    pattList.add(me._rexLingoFunctionEnd)
    replStrList.add(RETURN&"_global.$.debug.pop()"&RETURN&"end"  &RETURN)

    -- breakpoints
    pattList.add("--\{B\}")
    replStrList.add(me._breakPointLingo)

    tCode = me._regex.replace_multi(pattList, replStrList, tCode, "gmi")

  else

    -- JS
    pattList = [me._rexJsFunctionStart, me._rexJsFunctionEnd, "//\{B\}"]
    replStrList = ["$1"&me._dbgTry, me._dbgCatch&"}", me._breakPointJS]
    tCode = me._regex.replace_multi(pattList, replStrList, tCode, "gm")
    put "this.scriptfile='"&me._regex.replace("\\", "\\", tScriptFile)&"';" before tCode

  end if

  -- /DEBUG ##############################################################################

  m.scriptText = tCode

  -- create new instance
  instanceRef = m.script.rawNew()
  instanceRef.setProp(#$, me)
  instanceRef.setProp(#libdir, tLibFolder & me.pd)

  -- call constructor, if it exists
  h = instanceRef.handlers()
  if h.getPos(#new) then instanceRef.new(tArgs)

  -- add instanceRef to $
  me[tLibName] = instanceRef

  -- save instanceRef in libdict (folder=>instanceRef dict)
  me._.libdict.addProp(tLibFolder, instanceRef)

  -- return the instanceRef
  return instanceRef

end

----------------------------------------
-- Overwrites LSW's include(), adding debugging extensions
----------------------------------------
on include (me, tScriptFile, tForceType)

  isRelative = me.isRelativePath(tScriptFile)

  fp = xtra("fileIO").new()

  --if not isRelative then
  tScriptFile = me.sanitizePath(tScriptFile)
  --end if

  if isRelative then
    found = FALSE
    repeat with dir in _player.searchPathList
      fn = dir & tScriptFile

      -- script is already included!
      if not voidP(me._.scriptdict[fn]) then return me._.scriptdict[fn]

      -- does file exist?
      fp.openFile(fn, 1)
      found = (fp.status()=0)
      if found then
        tScriptFile = fn -- now absolute
        exit repeat
      end if
    end repeat

  else

    -- script is already included!
    if not voidP(me._.scriptdict[tScriptFile]) then return me._.scriptdict[tScriptFile]

    -- does file exist?
    fp.openFile(tScriptFile, 1)
    found = (fp.status()=0)
  end if

  if not found then return FALSE

  -- read contents of script file
  tCode = fp.readFile()
  fp.closeFile()
  fp = 0

  -- extract basename and type
  od = _player.itemdelimiter
  _player.itemdelimiter = me.pd
  tBasename = the last item of tScriptFile
  _player.itemdelimiter = "."
  tType = the last item of tBasename
  _player.itemdelimiter = od

  -- create new script
  m = new(#script, castlib("lswlib"))
  if symbolP(tForceType) then
    m.scriptType = tForceType
  else if tCode.line[1] starts "--!" OR tCode.line[1] starts "//!" then
    str = tCode.line[1].char[4..tCode.line[1].length]
    if ["movie", "parent", "score"].getPos(str) then
      m.scriptType = symbol(str)
    end if
  else
    m.scriptType = #movie -- default
  end if
  if tType="js" OR tType="jsw" then
    m.scriptSyntax = #javascript
  else -- assume ls or lsw
    m.scriptSyntax = #lingo
  end if
  m.name = tBasename

  -- DEBUG ###############################################################################

  if m.scriptSyntax=#lingo then

    pattList = ["\{\{1\}\}", "\{\{2\}\}", "\{\{4\}\}"]
    replStrList = [m.name&&"(#"&m.scriptType&")", tScriptFile, string((m.scriptType<>#movie)+1)]
    tDbgCode = me._regex.replace_multi(pattList, replStrList, me._dbgCode)

    -- if parent script, save object where error occured, so properties can be listed in alterHook
    if m.scriptType=#parent then
      put RETURN&"if _global.$.debug._dbgFlag then _global.$.debug._errObj=param(1)" after tDbgCode
    else
      put RETURN&"_global.$.debug._errObj=VOID" after tDbgCode
    end if

    -- add debug code to all hander beginnings
    tCode = me._regex.replace_callback(me._rexLingoFunctionStart, tCode, #replace, me, "gmi", tDbgCode)

		-- replace "return foo" with "return"&" _global.$.debug.r("&foo&")"
  	-- note: i doesn't work for "return" vs "RETURN"
    tCode = me._regex.replace(me._rexLingoReturn, "return"&" _global.$.debug.r($1)", tCode, "gm")
    
    pattList = []
    replStrList = []

    -- precede function end "end" width "_global.$.debug.pop()"
    pattList.add(me._rexLingoFunctionEnd)
    replStrList.add(RETURN&"_global.$.debug.pop()"&RETURN&"end"  &RETURN)

    -- breakpoints
    pattList.add("--\{B\}")
    replStrList.add(me._breakPointLingo)

    tCode = me._regex.replace_multi(pattList, replStrList, tCode, "gmi")

  else

    -- JS
    pattList = [me._rexJsFunctionStart, me._rexJsFunctionEnd, "//\{B\}"]
    replStrList = ["$1"&me._dbgTry, me._dbgCatch&"}", me._breakPointJS]
    tCode = me._regex.replace_multi(pattList, replStrList, tCode, "gm")
    put "this.scriptfile='"&me._regex.replace("\\", "\\", tScriptFile)&"';" before tCode

  end if

  -- /DEBUG ##############################################################################

  m.scriptText = tCode
  return m.script
end

----------------------------------------
-- Embeds script in castlib (if not available yet)
-- Creates new instance (if not available yet), puts it in me.<basename> (or me.<name>)
-- Returns instance
--
-- @param {string} tLibName
-- @param {any} [tArgs]
-- @return {instance|false}
----------------------------------------
on importAndForget (me, tLibName, tArgs)

  fp = xtra("fileIO").new()

  -- API change: "lib" now fixed!
  tLibFolder = "lib"& $.PD & tLibName

  found = false
  repeat with dir in _player.searchPathList
    fn = dir & tLibFolder

    -- does folder exist?
    fp.openFile(fn, 1)
    found = (fp.status()=-36)
    fp.closeFile() -- just in case it was a file
    if found then
      tLibFolder = fn
      exit repeat
    end if
  end repeat
  if not found then return false

  -- look for index file
  repeat with tType in ["ls", "js"]
    tScriptFile = tLibFolder & $.PD & "index." & tType
    fp.openFile(tScriptFile, 1)
    found = (fp.status()=0)
    if found then exit repeat
  end repeat
  if not found then return false -- no index file found!

  -- read and close file
  tCode = fp.readFile()
  fp.closeFile()
  fp = 0

  -- create new script
  m = new(#script, castlib("lswlib"))
  m.scriptType = #parent -- only parent allowed!
  if tType="js" OR tType="jsw" then
    m.scriptSyntax = #javascript
  else -- ls, lsw
    m.scriptSyntax = #lingo
  end if
  m.name = tLibName
  m.scriptText = tCode

  -- create new instance
  instanceRef = script(m.number).rawNew()
  instanceRef.setProp(#$, $)
  instanceRef.setProp(#libdir, tLibFolder & $.PD)

  -- call constructor, if it exists
  h = instanceRef.handlers()
  if h.getPos(#new) then instanceRef.new(tArgs)

  m.erase()

  -- return the instanceRef
  return instanceRef

end

----------------------------------------
-- @callback
----------------------------------------
on replace (me, args, tDbgCode)
  handlerLine = args[1] -- 2?
  handlerName = args[3]
  tDbgCode = me._regex.replace("\{\{3\}\}", handlerName, tDbgCode)
  return handlerLine&RETURN&tDbgCode
end

----------------------------------------
-- Shows lingo error infos
-- @callback
----------------------------------------
on alertHook (me, errStr, msg, type)
  case (type) of
    #script:

      -- stop adding to callstack
      me._dbgFlag = FALSE

      -- show error
      --err errStr&": "&msg
      str = "ERROR:" &$.LF
      put errStr&": "&msg &$.LF after str

      e = me._callStack.getlast()
      put "Function:" && e.handler &$.LF after str
      put "Arguments:" && e.params &$.LF after str
      put "Script:" && e.name && "("&e.file&")" &$.LF after str

      -- show properties
      if objectP(me._errObj) then
        put $.LF after str
        put "Properties:" &$.LF after str
        cnt = me._errObj.count
        repeat with i = 1 to cnt
          put "  "&me._errObj.getPropAt(i)&": "&me._errObj.getAt(i) &$.LF after str
        end repeat
      end if

      -- show call stack
      put $.LF after str
      put "Callstack:" &$.LF after str
      cnt = me._callStack.count
      last = 1 --max(1, cnt-me.stackShowMax+1)
      repeat with i = cnt down to last
        e = me._callStack[i]
        put "  "&e.handler && e.name && e.params &$.LF after str
      end repeat
			
			me._err(str)
			
      -- exit with errorCode 1
      if me.stopOnError then exitCode(1)

      -- otherwise continue to run
      me._dbgFlag = TRUE
      return -1

    otherwise:
      return 0

  end case
end

----------------------------------------
-- Shows JS error infos
----------------------------------------
on traceJsError(me, fn, e, stack, scriptfile, locals)

  str = "ERROR:" &$.LF
  
  put e.toString() &$.LF after str
  --put "Message:" && e.message &$.LF after str
  if not voidP(scriptfile) then put "Script:" && scriptfile &$.LF after str
  put "Function:" && fn &$.LF after str
  put "Line Number:" && e.lineNumber &$.LF after str

  -- show local vars
  put $.LF after str
  put "Local Variables:" &$.LF after str
  cnt = locals.count
  repeat with i = 1 to cnt
    put "  "&locals.getPropAt(i)&": "&locals[i] &$.LF after str
  end repeat

  -- show call stack
  put $.LF after str
  put "Stack:" &$.LF after str
  cnt = stack.length-1
  repeat with i = 0 to cnt
    put "  "&stack[i] &$.LF after str
  end repeat

	me._err(str)
	
  if me.stopOnError then exitCode(1)

end

----------------------------------------
-- triggered by Lingo/LSW breakpoint
----------------------------------------
on lingoBreak (me)

  -- stop adding to callstack
  me._dbgFlag = FALSE

  str = "BREAKPOINT:" &$.LF

  e = me._callStack.getlast()
  put "Function:" && e.handler &$.LF after str
  put "Arguments:" && e.params &$.LF after str
  put "Script:" && e.name && "("&e.file&")" &$.LF after str
	
  -- show properties
  if objectP(me._errObj) then
    put $.LF after str
    put "Properties:" &$.LF after str
    cnt = me._errObj.count
    repeat with i = 1 to cnt
    	v = me._errObj.getAt(i)
    	if voidP(v) then v = "<Void>"
      put "  "&me._errObj.getPropAt(i)&": "&v &$.LF after str
    end repeat
  end if

  -- show local vars
  put me._file.getString(me._traceFile) &$.LF after str
	
  -- show call stack
  put $.LF after str
  put "Callstack:" &$.LF after str
  cnt = me._callStack.count
  last = max(1, cnt-me.stackShowMax+1)
  repeat with i = cnt down to last
    e = me._callStack[i]
    put "  "&e.handler && e.name && e.params &$.LF after str
  end repeat

	me._err(str)
	
  _player.alert("PRESS TO CONTINUE")

  me._dbgFlag = TRUE
end

----------------------------------------
-- triggered by JS breakpoint
----------------------------------------
on jsBreak (me, fn, stack, scriptfile, locals)

  str = "BREAKPOINT:" &$.LF
   
  if not voidP(scriptfile) then put "Script:" && scriptfile &$.LF after str
  put "Function:" && fn &$.LF after str

  -- get line number from stack
  tmp = me._regex.split("@:", stack[1])
  put "Line Number:" && tmp.getLast() &$.LF after str

  -- show local vars
  put $.LF after str
  put "Local Variables:" &$.LF after str
  cnt = locals.count
  repeat with i = 1 to cnt
    put "  "&locals.getPropAt(i)&": "&locals[i] &$.LF after str
  end repeat

  -- show call stack
  put $.LF after str
  put "Stack:" &$.LF after str
  cnt = stack.length-1
  repeat with i = 1 to cnt
    put "  "&stack[i] &$.LF after str
  end repeat
  
	me._err(str)
	
  _player.alert("PRESS TO CONTINUE")
end

----------------------------------------
--
----------------------------------------
on _err (me, str)
	if not voidP($[#scite]) then
		err (str)
	else if not voidP($[#console]) then
		$.console.error(str)
	else
		put str
	end if
end