/**
 * ISO Maker 0.1
 *
 * Scriptfile encoding: UTF-8 (test: äöü)
 * Comments in JSDoc 3 format
 *
 * @file Core app file for ISO Maker.
 * @author Valentin Schmidt
 * @copyright Valentin Schmidt 2013
 * @version 0.1.1
 *
 * @todo no more folder tree creation in tmp ?
 * @todo add last mod dates to ISO ?
 */


//################################################
// REQUIRE
//################################################

// built-in
var fs = require('fs');
var qs = require('querystring');

// load from modules folder
var qt = require('node-qt');
var QT = require('node-qt/qt-constants'); // optional: used named constants
var ds = require('datastructures'); // for Tree + Node

//require('dbg');
//dbg(process.env);


//################################################
// GLOBALS
//################################################

var
	app_name = 'ISO Maker', /** @constant */
	app_version = '0.1', /** @constant */
	treeItemDefaultFlags = 61, /** @constant */
	illegalChars = ':/\\\\"*?<>|TAB', /** @constant */
	colCnt = 4, /** @constant */

	settings = {
	  confirm:true,       // show confirm dialog for delete operations?
	  emptyFolders:true,  // also create empty folders in ISO image (otherwise omitted)
	  useTmpFolder:false,  // use user's TMP folder instead of local subfolder data/tmp?
	  toolbar:true,       // show toolbar
	  statusbar:true,     // show statusbar

	  mkisofs_args: [ // uncomment or add your own
	  	'-J',												// Joliet File System
	  	
//  		'-hfs ',											// Apple HFS File System
//  		'-output-hfs-charset default',
  		
//  		'-udf',                     // UDF File System
//  		'-R',												// RockRidge File System (for Unix/Linux)

// 			'-dvd-video',               // Generate DVD-Video compliant UDF file system
//
//	  	'-volid ID',            		// Set Volume ID
//  		'-preparer PREP',     			// Set Volume preparer
//	  	'-publisher PUB',           // Set Volume publisher
//	  	'-sysid ID',                // Set System ID
//
//			'-verbose',                	// Verbose
//  		'-log-file LOG_FILE',       // Re-direct messages to LOG_FILE
//
//			'-exclude GLOBFILE',      	// Exclude file(s)
//
//			'-b boot.bin',							// Set El Torito boot image name
//			'-no-emul-boot',            // Boot image is 'no emulation' image
//			'-hide boot.bin',
//			'-hide boot.catalog'
		]
	},

	tmp_posix = (settings.useTmpFolder?process.env.TMP.replace(/\\/g, '/'):__dirname+'/tmp') + 
	  '/im'+Math.floor(Math.random()*(999999-100000+1)+100000),

	curFile='',
	isModified = false,

	row = 0,
	rowCnt = 10,

	folderTree,
	folderRootNode,
	currentFolderNode,

	rootItem,
	currentTreeItem,

	// qt widgets
	app,
	mainWindow,
	menuBar,
	toolBar,
	hsplitter,
	treeWidget,
	tableWidget,
	statusBar,
	statusBarLabel,

	tableWidgetItems = [],
	tableWidgetCellWidgets = [],
	actions = {},
	menus = {};


//################################################
// FUNCTIONS
//################################################

/**
 * Constructor for File objects
 *
 * @constructor
 * @this {File}
 * @param {string} name - The file's filename on CD.
 * @param {string} path - The file's original filesystem path.
 * @param {number} size - The file's size (in bytes).
 * @param {<Date>} mtime - Last modified Date
 */
function File(name, path, size, mtime){
	this.name = name;
	this.path = path;
	this.size = size;
	this.mtime = mtime;
}

/**
 * Sort function for File objects.
 *
 * @callback
 * @static
 */
File.sort = function(a, b) {
	a = (a.name+'').toLowerCase();
	b = (b.name+'').toLowerCase();
	return a<b?-1:(a>b?1:0);
}

/**
 * Constructor for FolderData objects
 *
 * @constructor
 * @this {FolderData}
 * @param {<Date>} mtime - Last modified Date
 * @param {number} size - The folder's curent total size (in bytes).
 */
function FolderData(mtime, size){
	this.mtime = mtime!=undefined?mtime:new Date();
	this.size = size!=undefined?size:0;
	this.files = [];
}

/**
 * Creates compact string representation for date object: "DD.MM.YYYY HH:MM"
 * (better alternative: YYYY-MM-DD HH:MM ?)
 *
 * @param {<Date>} date
 * @returns {string}
 */
function getDateStr(d){
	var s, date = '';
	s = d.getDate().toString();
	if (s.length<2) s='0'+s;
	date += s+'.';
	s = (d.getMonth()+1).toString();
	if (s.length<2) s='0'+s;
	date += s+'.';
	s = d.getFullYear().toString();
	date += s+' ';
	s = d.getHours().toString();
	if (s.length<2) s='0'+s;
	date += s+':';
	s = d.getMinutes().toString();
	if (s.length<2) s='0'+s;
	date += s;
	return date;
}

/**
 * Creates human friendly file size string
 *
 * @param {number} size
 * @returns {string}
 */
function getFileSizeStr(size) {
	if (size==0) return '0 B';
	var i = Math.floor( Math.log(size) / Math.log(1024) );
	return ( size / Math.pow(1024, i) ).toFixed(1) * 1 + ' ' + ['B', 'KB', 'MB', 'GB', 'TB'][i];
}

/**
 * Shows a confirm dialog.
 *
 * @param {string} txt
 * @returns {boolean}
 */
function confirm(txt){
	var msgBox = new qt.QMessageBox;
	msgBox.setIcon(4);
	msgBox.setWindowTitle(app_name);
	msgBox.setText(txt);
	msgBox.setStandardButtons(QT.STANDARDBUTTON.OK | QT.STANDARDBUTTON.CANCEL);
	msgBox.setDefaultButton(QT.STANDARDBUTTON.OK);
	ret = msgBox.exec();
	return (ret == QT.STANDARDBUTTON.OK);
}

/**
 * Adds a folder (including all files and subfolders) to the specified TreeWidget Item
 * and folderNode (recursive)
 *
 * @param {string} path
 * @param {<QTreeWidgetItem>} treeItem
 * @param {<Node>} folderNode
 */
function addFilesystemFolderToNode(path, treeItem, folderNode){
	var name = path.split('/').pop();

	var stats = fs.statSync(path);
	var childFolderNode = folderTree.addNode(folderNode.id, name, new FolderData(stats.mtime));

	// create treeWidgetItem
	var childTreeItem = new qt.QTreeWidgetItem;
	childTreeItem.setData(0,3, childFolderNode.id)
	childTreeItem.setText(0, name);
	childTreeItem.setFlags(treeItemDefaultFlags|2);
	childTreeItem.setIcon(0, __dirname, true);
	treeItem.addChild(childTreeItem);

	var flist = fs.readdirSync(path);
	flist.forEach(function(name){
		var p = path+'/'+name;
		var stats = fs.statSync(p);
		if (stats.isFile()){
			childFolderNode.data.files.push(new File(name, p, stats.size, stats.mtime));
			updateFolderSize(childFolderNode, stats.size);
		}else if(stats.isDirectory()){
			addFilesystemFolderToNode(p, childTreeItem, childFolderNode);
		}
	});
}

/**
 * Resets (empties) the tree widget and sets the root text to the specied name
 *
 * @param {string} name - Text of treeWidget's root item.
 */
function resetTreeWidget(name){
	treeWidget.setCurrentItem(rootItem);
	rootItem.setText(0, name);
	currentTreeItem = rootItem;
	var cnt = rootItem.childCount();
	for (i=cnt-1;i>=0;i--){
		node = rootItem.child(i);
		rootItem.removeChild(node);
		node.destroy();
	}
}

/**
 * Resets (empties) the table widget
 */
function resetTableWidget(){
	var w;
	for (var i=0;i<row;i++){
		w = tableWidget.cellWidget(i, 4);
		tableWidget.removeCellWidget(i, 4);
		w.destroy();
		w = tableWidget.cellWidget(i, 5);
		tableWidget.removeCellWidget(i, 5);
		w.destroy();
	}
	tableWidgetItems.forEach(function(item){
		item.destroy();
	});
	tableWidgetItems = [];
	row = 0;
}

/**
 * Displays the current folderNode in the table widget
 * (after sorting its subfolders and files alphabetically by name)
 */
function showCurrentNodeInTable(){
	resetTableWidget();

	folderTree.sortNodesByName(currentFolderNode.id); // sort subfolder
	currentFolderNode.data.files.sort(File.sort); // sort files

	var cnt = currentFolderNode.childNodes.length + currentFolderNode.data.files.length;
	if (cnt>rowCnt){
		rowCnt = cnt;
		tableWidget.setRowCount(rowCnt);
		tableWidget.resizeRowsToContents();
	}

	currentFolderNode.childNodes.forEach(function(node){
		addTableRowFolder(node);
	});
	currentFolderNode.data.files.forEach(function(f){
		addTableRowFile(f);
	});

	tableWidget.setCurrentCell(-1, -1); // no cell selected anymore

	var totalSize = folderRootNode.data.size;
	statusBarLabel.setText('Total Size: '+(totalSize<1024?totalSize+' Bytes':getFileSizeStr(totalSize)+' ('+totalSize+' Bytes)'));
}

/**
 * Creates a table widget item for the specified table widget cell
 *
 * @param {number} row
 * @param {number} col
 * @param {string} str - Cell text.
 * @param {number} icon - path to icon image file (or arbitrary file, if iconFromFile=true).
 * @param {boolean} iconFromFile - if true, try to get system icon for file specified as icon parameter.
 */
function setTableCell(row, col, str, icon, iconFromFile){
	var tableWidgetItem = new qt.QTableWidgetItem;
	tableWidgetItem.setText(str);
	if (icon) tableWidgetItem.setIcon(icon, iconFromFile);
	tableWidget.setItem(row, col, tableWidgetItem);
	tableWidgetItems.push(tableWidgetItem); // saved for later destruction
}

/**
 * Creates a table widget row for the specified folderNode
 *
 * @param {<Node>} folderNode
 */
function addTableRowFolder(folderNode){
	setTableCell(row, 0, folderNode.name, __dirname, true); // get system folder icon
	setTableCell(row, 1, 'Folder');
	setTableCell(row, 2, folderNode.data.size);
	setTableCell(row, 3, getDateStr(folderNode.data.mtime));

	row++;
}

/**
 * Creates a table widget row for the specified File instance
 *
 * @param {<File>} f
 */
function addTableRowFile(f){
	setTableCell(row, 0, f.name, f.path, true); // get system/explorer icon for this file
	setTableCell(row, 1, f.path);
	setTableCell(row, 2, f.size);
	setTableCell(row, 3, getDateStr(f.mtime));

	row++;
}

/**
 * Handles 'New' action
 */
function newFile(){
  if (maybeSave()) {

  	// RESET
		folderTree = new ds.Tree('CD');
		folderRootNode  = folderTree.getRootNode();
		folderRootNode.data = new FolderData();
		currentFolderNode = folderRootNode;

		resetTreeWidget('CD');
		resetTableWidget();

   	setCurrentFile('');
  }
}

/**
 * Handles 'Open' action
 */
function open(){
  if (maybeSave()){
		var fn = qt.QFileDialog.getOpenFileName('', '', '*.ism'); // caption, dir, filter, *options
		if (fn) loadFile(fn);
	}
}

/**
 * Handles 'Save' action
 *
 * @param {boolean} sync
 * @returns {boolean} Success
 */
function save(){
  if (!curFile){
		return saveAs();
  } else {
		return saveFile(curFile);
  }
}

/**
 * Handles 'SaveAs' action
 *
 */
function saveAs(){
	// caption, dir, filter, selectedFilter, options
	var fn = qt.QFileDialog.getSaveFileName('', rootItem.text(0)+'.ism', '*.ism'); //curFile
	if (!fn) return false;
	return saveFile(fn);
}

/**
 * If applicable, asks if changes should be saved before proceeding
 *
 * @returns {boolean} Returns true if process can be proceeded
 */
function maybeSave(){
  if (isModified){
  	var msgBox = new qt.QMessageBox;
  	msgBox.setIcon(4);
  	msgBox.setWindowTitle(app_name);
  	msgBox.setText('The document has been modified.\nDo you want to save your changes?');
  	msgBox.setStandardButtons(QT.STANDARDBUTTON.SAVE | QT.STANDARDBUTTON.DISCARD | QT.STANDARDBUTTON.CANCEL);
 		msgBox.setDefaultButton(QT.STANDARDBUTTON.SAVE);
 		ret = msgBox.exec();
    if (ret == QT.STANDARDBUTTON.SAVE)
			return save();
    else if (ret == QT.STANDARDBUTTON.CANCEL)
			return false;
  }
  return true;
}

/**
 * Loads file/folder list from specified *.ISM file.
 *
 * @param {string} fn
 */
function loadFile(fn){
	fs.readFile(fn, 'binary', function (err, data) {
	  if (err) {
	    //throw err;
	    qt.QMessageBox.warning('ISOMaker', 'Cannot read file '+fn);
	    return;
	  }

	  var lines = data.split('\n');

	  header = lines.shift();
	  if (header.substr(0,8)!='ISOMAKER'){
	  	qt.QMessageBox.warning('ISOMaker', 'Not an ISOMaker file: '+fn);
	  	// TODO: version check?
	  	return;
	  }

	  if (lines[lines.length-1]=='') lines.pop();
	  var volumeName = lines.shift();

  	// RESET
		folderTree = new ds.Tree('CD');
		folderRootNode  = folderTree.getRootNode();
		folderRootNode.data = new FolderData();
		currentFolderNode = folderRootNode;

		resetTreeWidget(volumeName);
		resetTableWidget();

		// locals
	  var stats;
	  var mtime;
		var parts;
		var treePath;
		var treePathParts;
		var childTreeItem;
		var nameInImage;
		var is_folder;
		var node;

  	var treeItemTable = {}; // loockup table of all tree idems: id->item
  	treeItemTable[folderRootNode.id] = rootItem;

	  lines.forEach(function(l){

	  	// parse current line
	  	is_folder = l.charAt(0)=='#';

	  	if (is_folder){
	  		l = l.substr(1);
	  		if (l.charAt(0)=='/') l = l.substr(1);
	  		parts = l.split('|');

	  		treePath = parts[0];
	  		mtime = new Date(parts[1]);
	  		 
	  		treePathParts = treePath.split('/');

	  		// loop through folder names
	  		node = folderTree.addNodeByPath(treePathParts, function(newNode){
	  			newNode.data = new FolderData();
					//create widgetItem for new node
					childTreeItem = new qt.QTreeWidgetItem;
					childTreeItem.setData(0,3, newNode.id);
					childTreeItem.setText(0, newNode.name);
					childTreeItem.setFlags(treeItemDefaultFlags|2);
					childTreeItem.setIcon(0, __dirname, true);
					treeItemTable[newNode.id] = childTreeItem;
					treeItemTable[newNode.parentID].addChild(childTreeItem);
	  		});
	  		node.data = new FolderData(mtime);

			}else{// IS FILE
		  	parts = l.split('=');
		  	treePath = parts[0];
		  	if (treePath.charAt(0)=='/') treePath = treePath.substr(1);
		  	treePathParts = treePath.split('/');

		  	path = parts[1];
		  	nameInImage = treePathParts.pop();

				node = treePathParts.length==0? folderRootNode : folderTree.addNodeByPath(treePathParts, function(newNode){
					newNode.data = new FolderData();
					//create widgetItem for new node
					childTreeItem = new qt.QTreeWidgetItem;
					childTreeItem.setData(0,3, newNode.id);
					childTreeItem.setText(0, newNode.name);
					childTreeItem.setFlags(treeItemDefaultFlags|2);
					childTreeItem.setIcon(0, __dirname, true);
					treeItemTable[newNode.id] = childTreeItem;
					treeItemTable[newNode.parentID].addChild(childTreeItem);
	  		});

				if (fs.existsSync(path)){
	  			stats = fs.statSync(path);
	  			updateFolderSize(node, stats.size);
	  		}else{
	  			// SHOW ALERT!
	  			stats = {size:0, mtime:0};
	  		}
				node.data.files.push(new File(nameInImage, path, stats.size, stats.mtime));
			}

		}); // end lines.forEach(function(l){

	  showCurrentNodeInTable();
	  setCurrentFile(fn);

	});//end fs.readFile(
}

/**
 * Saves current file/folder list as specified *.ISM file.
 * Returns true if successful, false on error
 *
 * @param {string} fn - Full file path
 * @returns {boolean}
 */
function saveFile(fn){
  var str = 'ISOMAKER '+app_version+'\n'+ folderRootNode.name+'\n';
  str += getFolderListing(folderRootNode, '/');
	try{
		fs.writeFileSync(fn, str, 'binary');
	}catch(e){
		return false;
	}
	setCurrentFile(fn);
	statusBar.showMessage('File saved.', 5000);
	return true;
}

/**
 * Exports current file/folder list as ISO image.
 *
 * @param {string} fn - Full file path
 */
function makeISO(fn){

	var clearTmp = function(){
		if (!fs.existsSync(tmp_posix)) return;
		var files = fs.readdirSync(tmp_posix);
		files.forEach(function(f, i){
		  if (fs.statSync(tmp_posix+'/'+f).isFile()) fs.unlinkSync(tmp_posix+'/'+f);
		});
		deleteFolder(tmp_posix);
	}
	
	clearTmp(); // delete
	fs.mkdirSync(tmp_posix); // (re)create

	// create tmp files
	var str = getList(folderRootNode, '/');
	try{
		fs.writeFileSync(tmp_posix+'/~files.txt', str, 'binary');
	}catch(e){
		return false;
	}

	// create folder tree
	fs.mkdirSync(tmp_posix+'/root/');

	createFolder(folderRootNode, tmp_posix+'/root/', true);

	var more_args = settings.mkisofs_args.length?settings.mkisofs_args.join(' ')+' ':'';

	var cmd = '"'+__dirname+'\\bin\\mkisofs" '+
'-graft-points '+
more_args+
'-V "'+folderRootNode.name+'" '+
'-o "'+fn+'" '+
'-path-list "'+tmp_posix+'/~files.txt" '+
'"'+tmp_posix+'/root"';

	//console.log(cmd);

	var exec = require('child_process').exec;
	var aborted = false;

	progressDialog = new qt.QProgressDialog;
	progressDialog.setWindowModality(QT.WINDOWMODALITY.APPLICATIONMODAL);
	progressDialog.setWindowTitle("Creating ISO file...");
	progressDialog.setFixedWidth(400);
	progressDialog.setLabelText(''); // Progress:
	progressDialog.setCancelButtonText('Abort'); // Abort
	progressDialog.setMinimum(0);
	progressDialog.setMaximum(100);

	progressDialog.on("canceled", function(){
		console.log("++++++++++ ISO EXPORT CANCELED ++++++++++");
		aborted = true;
		clearInterval(intervalId);
		child.kill(); // = cmd.exe, not mkisofs.exe!
		exec('taskkill /F /IM mkisofs.exe', function (error, stdout, stderr){
			// clean up -> file still locked!?!
			setTimeout(function(){
				if (fs.existsSync(fn)) fs.unlinkSync(fn);
				clearTmp();
			},2000);// TODO: interval, repeat until succesfull
		});
	});

	//progressDialog.on("closeEvent", quit); //process.exit);
	progressDialog.show();

	var intervalId = setInterval(function(fn){
		if (folderRootNode.data.size && fs.existsSync(fn)){
			var stats = fs.statSync(fn);
			progressDialog.setValue(100*stats.size/folderRootNode.data.size);
		}
	}, 1000, fn); // update progress every 1000 ms

	var child = exec(cmd, function (error, stdout, stderr){
		clearInterval(intervalId);
		progressDialog.setValue(100);
		console.log('stdout: ' + stdout);
		console.log('stderr: ' + stderr);
		if (error !== null) console.log('exec error: ' + error);

		statusBar.showMessage(aborted?
			'Export aborted' :
			'ISO File successfully saved',
			5000);
		progressDialog.hide();
		clearTmp();
	});
}

/**
 * Renames currently selected table row.
 */
function renameItem(){
	var row = tableWidget.currentRow();
	var is_folder = row<currentFolderNode.childNodes.length;
	var item = tableWidget.item(row, 0);
	var old_name = item.text();
	var res = qt.QInputDialog.getText('Rename '+(is_folder?'folder':'file'),
	    'Enter a new name:', old_name);
	if (typeof res!='string' || res=='' || res==old_name) return;
	item.setText(res);
	if (is_folder){
		// change widgetTreeItem
		var treeItem = currentTreeItem.child(row);
		treeItem.setText(0, res);
		var itemID = treeItem.data(0,3);
		// change in folderTree
		folderTree.getNode(itemID).name = res;
	}else{
		currentFolderNode.data.files[row-currentFolderNode.childNodes.length].name = res;
	}
	showCurrentNodeInTable();
	treeWidget.sortItems(0, 0);
	isModified = true;
}

/**
 * Removes currently selected table row.
 */
function removeTableWidgetRow(){
	var row = tableWidget.currentRow();
	var is_folder = row<currentFolderNode.childNodes.length;
	if (settings.confirm){
		var ok = confirm('Remove selected '+(is_folder?'folder':'file')+'?');
		if (!ok) return;
	}
	//remove table row
	if (is_folder){
		// remove from treeWidget
		var treeItem = currentTreeItem.child(row);
		currentTreeItem.removeChild(treeItem);
		treeItem.destroy();
		var childNode = currentFolderNode.childNodes[row];
		// remove from folderTree
		folderTree.deleteNode(childNode.id);
		// calculate removed size
		updateFolderSize(currentFolderNode, -childNode.data.size);
	}else{
		var i = row-currentFolderNode.childNodes.length;
		updateFolderSize(currentFolderNode, -currentFolderNode.data.files[i].size);
		currentFolderNode.data.files.splice(i, 1);
	}
	showCurrentNodeInTable();
	isModified = true;
}

/**
 * Removes currently selected TreeWidget item row.
 */
function removeTreeWidgetItem(){
	var itemID = currentTreeItem.data(0,3);
	if (itemID==folderRootNode.id) return;
	if (settings.confirm){
		var ok = confirm('Remove selected folder?');
		if (!ok) return;
	}
	var p = folderTree.getParentNode(currentFolderNode.id);
	folderTree.deleteNode(currentFolderNode.id);
	updateFolderSize(p, -currentFolderNode.data.size);
	currentFolderNode = folderRootNode;

	var item = currentTreeItem;
	currentTreeItem = rootItem;
	treeWidget.setCurrentItem(rootItem); // ???
	item.destroy();

	isModified = true;
}

/**
 * Gets file/folder listing as ISM string for specified folderNode (recursive).
 *
 * @param {<Node>} fn
 * @param {string} p - Path that is prepended to folderNode's name
 * @returns {string}
 */
function getFolderListing(folderNode, p){
	var str = '';
	folderNode.childNodes.forEach(function(childNode){
		str += '#'+p+childNode.name+ '|'+ childNode.data.mtime.toGMTString() + '\n';
		str += getFolderListing(childNode, p+childNode.name+'/');
	});
	folderNode.data.files.forEach(function(f){
		str += p+f.name+'='+f.path+'\n';
	});
	return str;
}

/**
 * Updates size of folderNode (recursive).
 *
 * @param {<Node>} folderNode
 * @param {number} ds - size difference
 */
function updateFolderSize(folderNode, ds){
	folderNode.data.size += ds;
	var p = folderTree.getParentNode(folderNode.id);
	if (p) updateFolderSize(p, ds);	
}

/**
 * Deletes folder and all subfolders (recursive).
 * Doesn't delete files, and therefor doesn't delete any folders containing files.
 *
 * @param {string} path
 */
function deleteFolder(path) {
  if( fs.existsSync(path) ) {
    var files = fs.readdirSync(path);
    files.forEach(function(f, i){
        var curPath = path + '/' + f;
        if(fs.statSync(curPath).isDirectory()) { // recurse
					deleteFolder(curPath);
        }
    });
    fs.rmdirSync(path);
  }
};

/**
 * Creates a folder and all subfolders (recursive).
 *
 * @param {<Node>} folderNode
 * @param {string} p - Path that is prepended to folderNode's name
 * @param {boolean} [omit] - If true, omit outer node, proceed with child nodes
 */
function createFolder(folderNode, p, omit){
	if (!omit){
		fs.mkdirSync(p+folderNode.name);
		p+=folderNode.name+'/';
	}
	folderNode.childNodes.forEach(function(childNode){
		createFolder(childNode, p);
	});
}

/**
 * Gets file listing in mkisofs format for specified folderNode (recursive).
 *
 * @param {<Node>} folderNode
 * @param {string} p - Path that is prepended to folderNode's name
 * @returns {string}
 */
function getList(folderNode, p){
	var str = '';
	folderNode.data.files.forEach(function(f){
		//str += '"'+p+f.name+'"="'+f.path+'"\n';
		str += p+f.name+'='+f.path+'\n';
	});
	folderNode.childNodes.forEach(function(childNode){
		str += getList(childNode, p+childNode.name+'/');
	});
	return str;
}

/**
 * Sets internally stored 'current file' to given file.
 *
 * @param fn {String}
 */
function setCurrentFile(fn){
	curFile = fn;
	isModified = false;
}

/**
 * Handles Exit action.
 *
 * @callback
 */
function exit(e){
  if (maybeSave()) {
		process.exit();
  } else if (e) e.ignore();
}


//################################################
// START
//################################################

// CREATE APP AND WIDGETS
var app = new qt.QApplication;

var mainWindow = new qt.QMainWindow(true);
var menuBar = new qt.QMenuBar;
var toolBar = new qt.QToolBar;
var hsplitter = new qt.QSplitter;
var treeWidget = new qt.QTreeWidget(0); // 1
var tableWidget = new qt.QTableWidget(0); // 1
var statusBar = new qt.QStatusBar;
var statusBarLabel = new qt.QLabel;
statusBar.addWidget(statusBarLabel, 1);

// TREE DATA STRUCTURE
folderTree = new ds.Tree('CD');
folderRootNode  = folderTree.getRootNode();
folderRootNode.data = new FolderData();
currentFolderNode = folderRootNode;


//################################################
// CONFIGURE MENUBAR
//################################################

// MENU FILE
menus.menu_file = menuBar.addMenu('&File');

// FILE:NEW
actions.action_new = menus.menu_file.addAction('&New').on('triggered', newFile);
actions.action_new.setIcon(__dirname + '/assets/toolbar/new.png');
actions.action_new.setShortcut('Ctrl+N');
actions.action_new.setStatusTip('Create a new file');

// FILE:OPEN
actions.action_open = menus.menu_file.addAction('&Open').on('triggered', open);
actions.action_open.setIcon(__dirname + '/assets/toolbar/open.png');
actions.action_open.setShortcut('Ctrl+O');
actions.action_open.setStatusTip('Open an existing file');

// FILE:SAVE
actions.action_save = menus.menu_file.addAction('&Save').on('triggered', save);
actions.action_save.setIcon(__dirname + '/assets/toolbar/save.png');
actions.action_save.setShortcut('Ctrl+S');
actions.action_save.setStatusTip('Save the document to disk');

// FILE:SAVEAS
actions.action_saveas = menus.menu_file.addAction('Save &As...').on('triggered', saveAs);
actions.action_saveas.setStatusTip('Save the document under a new name');

menus.menu_file.addSeparator();

actions.action_iso = menus.menu_file.addAction('Make ISO').on('triggered', function(){
	//if (maybeSave()) {
		var fn = qt.QFileDialog.getSaveFileName('', rootItem.text(0)+'.iso', '*.iso'); //curFile
		if (!fn) return false;
		makeISO(fn);
	//}
});
actions.action_iso.setShortcut('Ctrl+M');
actions.action_iso.setStatusTip('Create a (hybrid) ISO image');

menus.menu_file.addSeparator();

// FILE:EXIT
actions.action_exit = menus.menu_file.addAction('E&xit');
actions.action_exit.on('triggered', exit);
actions.action_exit.setShortcut('Ctrl+Q');
actions.action_exit.setStatusTip('Exit the application');

// MENU VIEW

menus.menu_view = menuBar.addMenu('&View');

// VIEW:SHOWTOOLBAR
actions.action_showtoolbar = menus.menu_view.addAction('Show ToolBar');
actions.action_showtoolbar.setCheckable(true);
actions.action_showtoolbar.setChecked(settings.toolbar);
actions.action_showtoolbar.on('toggled', function(checked){
	toolBar.setVisible(checked);
});

// VIEW:SHOWSTATUSBAR
actions.action_showstatusbar = menus.menu_view.addAction('Show StatusBar');
actions.action_showstatusbar.setCheckable(true);
actions.action_showstatusbar.setChecked(settings.statusbar);
actions.action_showstatusbar.on('toggled', function(checked){
	statusBar.setVisible(checked);
});

// MENU HELP
menus.menu_help = menuBar.addMenu('&Help');

actions.action_about = menus.menu_help.addAction('About').on('triggered', function(){
	qt.QMessageBox.about('About', '<b>'+app_name+' v'+app_version+'</b><br><br>Simple mkisofs GUI for creating ISO images that can be burned to CD/DVD with any burning software.<br><br>This application is based on <a href="http://nodejs.org/">Node.js</a>, <a href="http://qt-project.org/doc/qt-4.8/">Qt 4.8</a> and a customized and extended version of the <a href="https://github.com/arturadib/node-qt#readme">Node-Qt Addon</a> and was written in pure <a href="https://en.wikipedia.org/wiki/JavaScript">JavaScript</a>.<br><br>License: <a href="http://en.wikipedia.org/wiki/WTFPL">WTFPL 2.0</a>');
});
menus.menu_help.addSeparator();
// SHOW A CHM HELP FILE
actions.action_help = menus.menu_help.addAction('About Node.js').on('triggered', function(){
	var exec = require('child_process').exec;
	exec('start "" "'+__dirname+'\\assets\\help\\Node.js-0.10.15.chm"', function (error, stdout, stderr){});
});
actions.action_aboutqt = menus.menu_help.addAction('About Qt').on('triggered', function(){
	qt.QMessageBox.aboutQt('About Qt');
});


if (process.execArgv.indexOf('--expose-gc')>-1){
	menus.menu_help.addSeparator();
	actions.gc = menus.menu_help.addAction('Force Garbage Collection').on('triggered', function(){
		global.gc();
	});
}


//################################################
// CONFIGURE TOOLBAR
//################################################

toolBar.setWindowTitle('ToolBar');
toolBar.on('visibilityChanged', function(vis){
	// update menu
	actions.action_showtoolbar.setChecked(vis);
});
toolBar.addAction(actions.action_new);
toolBar.addAction(actions.action_open);
toolBar.addAction(actions.action_save);


//################################################
// CONFIGURE TREEWIDGET
//################################################

treeWidget.setHeaderHidden(true);

treeWidget.on('currentItemChanged', function(newItem){
	currentTreeItem = newItem;
	var currentTreeItemID = currentTreeItem.data(0,3);
	currentFolderNode = folderTree.getNode(currentTreeItemID);
	showCurrentNodeInTable();
});

treeWidget.on('itemChanged', function(item, col){
	var itemID = item.data(0,3);
	var newName = item.text();

	// TODO: CHECK/REMOVE ILLEGAL CHARS

	folderTree.getNode(itemID).name = newName;
});

// CREATE CONTEXT MENU
treeWidget.contextMenu = new qt.QMenu();

actions.action_subfolder = treeWidget.contextMenu.addAction('New Subfolder...');
actions.action_subfolder.on('triggered', function(){
	var name = qt.QInputDialog.getText('New Folder', 'Enter name:', '');
	if (typeof name!='string' || name=='') return;
	name = name.replace(new RegExp('['+illegalChars+']', 'g'), '');
	// creade folderNode
	var childFolderNode = folderTree.addNode(currentFolderNode.id,
	    name, new FolderData());
	// create treeWidgetItem
	var childTreeItem = new qt.QTreeWidgetItem;
	childTreeItem.setData(0,3, childFolderNode.id)
	childTreeItem.setText(0, name);
	childTreeItem.setFlags(treeItemDefaultFlags|2);
	childTreeItem.setIcon(0, __dirname, true);
	currentTreeItem.addChild(childTreeItem);

	showCurrentNodeInTable();
	isModified = true;
});

treeWidget.contextMenu.addSeparator();

actions.action_treeitem_remove = treeWidget.contextMenu.addAction('Remove');
actions.action_treeitem_remove.on('triggered', removeTreeWidgetItem);


treeWidget.setContextMenuPolicy(QT.CONTEXTMENUPOLICY.CUSTOMCONTEXTMENU); // 3
treeWidget.on('customContextMenuRequested', function(p){
	var p1 = mainWindow.geometry(),
		p2 = hsplitter.geometry(),
		p3 = treeWidget.geometry();
	var x = p1.x()+p2.x()+p3.x()+p.x();
	var y = p1.y()+p2.y()+p3.y()+p.y();
	treeWidget.contextMenu.exec(x, y);
});

rootItem = new qt.QTreeWidgetItem;
rootItem.setData(0,3, folderRootNode.id);
rootItem.setText(0, folderRootNode.name);
rootItem.setFlags(treeItemDefaultFlags|2);
rootItem.setIcon(0, __dirname + '/assets/icons/cd.png');
treeWidget.addTopLevelItem(rootItem);

currentTreeItem = rootItem;


//################################################
// CONFIGURE TABLEWIDGET
//################################################

tableWidgetItems = [];
tableWidgetCellWidgets = [];

tableWidget.setColumnCount(colCnt+1);
tableWidget.setRowCount(rowCnt);
tableWidget.setAlternatingRowColors(true);
tableWidget.setHorizontalHeaderLabels(['Name','Path','Size','Modified','']);
tableWidget.setVerticalHeaderVisible(false);
tableWidget.setHorizontalStretchLastSection(true);
tableWidget.setHorizontalHeaderDefaultAlignment(1);
tableWidget.resizeRowsToContents();
tableWidget.setSelectionBehavior(1); // QAbstractItemView::SelectRows
tableWidget.setSelectionMode(1);     // QAbstractItemView::SingleSelection
tableWidget.setStyleSheet("QCheckBox{margin-left:10px;}");

// 470
tableWidget.setColumnWidth(0,120);
tableWidget.setColumnWidth(1,320);
tableWidget.setColumnWidth(2,60);
tableWidget.setColumnWidth(3,100);

// make non-editable
tableWidget.setEditTriggers(0);


// CREATE CONTEXT MENU
tableWidget.contextMenu = new qt.QMenu(); // tableWidget

actions.action_rename = tableWidget.contextMenu.addAction('Rename...');
actions.action_rename.on('triggered', renameItem);

tableWidget.contextMenu.addSeparator();

actions.action_remove = tableWidget.contextMenu.addAction('Remove');
actions.action_remove.on('triggered', removeTableWidgetRow);

tableWidget.setContextMenuPolicy(3);
tableWidget.on('customContextMenuRequested', function(p){
	var row = tableWidget.currentRow();
	if (row<currentFolderNode.childNodes.length+currentFolderNode.data.files.length){
		var p1 = mainWindow.geometry(),
			p2 = hsplitter.geometry(),
			p3 = tableWidget.geometry();
		var x = p1.x()+p2.x()+p3.x()+p.x();
		var y = p1.y()+p2.y()+p3.y()+p.y()+20;
		tableWidget.contextMenu.exec(x, y);
	}
});

tableWidget.on('cellDoubleClicked', function(row, col){
	var item = tableWidget.item(row, col);
	if (col!=0 || item==undefined) return;
	var is_folder = row<currentFolderNode.childNodes.length;
	var old_name = item.text();
	var res = qt.QInputDialog.getText('Rename '+(is_folder?'folder':'file'),
	    'Enter a new name:', old_name);
	if (typeof res!='string' || res=='' || res==old_name) return;
	res = res.replace(new RegExp('['+illegalChars+']', 'g'), '');
	item.setText(res);
	isModified = true;
	if (is_folder){
		// change widgetTreeItem
		var treeItem = currentTreeItem.child(row);
		treeItem.setText(0, res);
		var itemID = treeItem.data(0,3);
		// change in folderTree
		folderTree.getNode(itemID).name = res;
	}else{
		currentFolderNode.data.files[row-currentFolderNode.childNodes.length].name = res;
	}
	showCurrentNodeInTable();
	treeWidget.sortItems(0, 0);
});


//################################################
// CONFIGURE QSPLITTER
//################################################

hsplitter.setOrientation(1);
hsplitter.addWidget(treeWidget);
hsplitter.addWidget(tableWidget);
hsplitter.setCollapsible(0, true);
hsplitter.setCollapsible(1, false);
hsplitter.setStretchFactor(0, 0);
hsplitter.setStretchFactor(1, 1);
hsplitter.setSizes([160,480]);


//################################################
// CONFIGURE MAINWINDOW
//################################################

mainWindow.setAcceptDrops(true);
mainWindow.setWindowIcon(__dirname + '/assets/icons/app.png');
mainWindow.setMenuBar(menuBar);
mainWindow.addToolBar(toolBar);
toolBar.setVisible(settings.toolbar);
mainWindow.resize(800, 480);
mainWindow.setCentralWidget(hsplitter);
mainWindow.setStatusBar(statusBar);
statusBar.setVisible(settings.statusbar);
mainWindow.setWindowState(settings.state);
mainWindow.setWindowTitle(app_name);

mainWindow.on('closeEvent', exit);

mainWindow.on('keyPressEvent', function(e){
	if (e.key()==QT.KEY.KEY_DELETE){
		var w = app.focusWidget();
		var name = w.className();
		if (name=='QTableWidget'){
			removeTableWidgetRow();
		}	else if (name=='QTreeWidget' && currentTreeItem!=rootItem){
			removeTreeWidgetItem();
		}
	}
});

mainWindow.on('dropEvent', function(e){
	var data = e.mimeData();
	if (!data['text/uri-list']) return;
	var flist = data['text/uri-list'].toString().split('\r\n');
	flist.pop();
	var overwriteAll = false;

	flist.forEach(function(path){
		if (path.substr(0,8)=='file:///') path = path.substr(8);
		path = qs.unescape(path);
		var stats = fs.statSync(path);
		var name = path.split('/').pop();

		if (stats.isFile()){

			// check if unique
			var unique = true;
			for (var i in currentFolderNode.data.files){
				if (currentFolderNode.data.files[i].name==name){
					var overwrite = overwriteAll;
					if (!overwriteAll){
				  	var msgBox = new qt.QMessageBox;
				  	msgBox.setIcon(4);
				  	msgBox.setWindowTitle(app_name);
				  	msgBox.setText('A file named "'+name+'" already exists!\nOverwrite existing with new file(s)?');
				  	msgBox.setStandardButtons(QT.STANDARDBUTTON.YES | QT.STANDARDBUTTON.YESTOALL | QT.STANDARDBUTTON.NO);
				 		msgBox.setDefaultButton(QT.STANDARDBUTTON.NO);
				 		var ret = msgBox.exec();
						overwriteAll = (ret==QT.STANDARDBUTTON.YESTOALL);
 						overwrite = overwriteAll || (ret==QT.STANDARDBUTTON.YES);
 					}
 					if (overwrite){
 						updateFolderSize(currentFolderNode, -currentFolderNode.data.files[i].size);
 						currentFolderNode.data.files.splice(i, 1);
 					}else{
						unique = false;
					}
					break;
				}
			}
			if (unique){
				currentFolderNode.data.files.push(new File(name, path, stats.size, stats.mtime));
				updateFolderSize(currentFolderNode, stats.size);
				isModified = true;
			}

		}else if(stats.isDirectory()){

			// check if unique
			var unique = true;
			for (var i in currentFolderNode.childNodes){
				if (currentFolderNode.childNodes[i].name==name){
					var overwrite = overwriteAll;
						if (!overwriteAll){
				  	var msgBox = new qt.QMessageBox;
				  	msgBox.setIcon(4);
				  	msgBox.setWindowTitle(app_name);
				  	msgBox.setText('A folder named "'+name+'" already exists!\nReplace existing folder (and all contents) with new folder(s)?');
				  	msgBox.setStandardButtons(QT.STANDARDBUTTON.YES | QT.STANDARDBUTTON.YESTOALL | QT.STANDARDBUTTON.NO);
				 		msgBox.setDefaultButton(QT.STANDARDBUTTON.NO);
				 		var ret = msgBox.exec();
 						overwriteAll = (ret==QT.STANDARDBUTTON.YESTOALL);
 						overwrite = overwriteAll || (ret==QT.STANDARDBUTTON.YES);
 					}
 					if (overwrite){
 						var childNode = currentFolderNode.childNodes[i];
						// remove from folderTree
						folderTree.deleteNode(childNode.id);
						updateFolderSize(currentFolderNode, -childNode.data.size);
						// remove from treeWidget
						var treeItem = currentTreeItem.child(i);
						currentTreeItem.removeChild(treeItem);
						treeItem.destroy();
 					}else{
						unique = false;
					}
					break;
				}
			}
			if (unique){
				addFilesystemFolderToNode(path, currentTreeItem, currentFolderNode);
				isModified = true;
			}
		}
	});

	showCurrentNodeInTable();
	treeWidget.sortItems(0, 0);
});

// SHOW
mainWindow.show();

// HANDLE ARGUMENTS
if (process.argv.length>2){
	var f = process.argv[2];
	try{loadFile(f);}catch(e){f=false;}
}

// JOIN NODE'S EVENT LOOP
setInterval(app.processEvents, 0);

		foo = function(x){
			;
		}