<?php 
// RTF Generator Class
//
// VS: added UTF-8 support to addText
//
// Example of use:
// 	$rtf = new RTF();
// 	$rtf->setPaperSize(5);
// 	$rtf->setPaperOrientation(1);
// 	$rtf->setDefaultFontFace(0);
// 	$rtf->setDefaultFontSize(24);
// 	$rtf->setAuthor("noginn");
// 	$rtf->setOperator("me@noginn.com");
// 	$rtf->setTitle("RTF Document");
// 	$rtf->addColour("#000000");
// 	$rtf->addText($_POST['text']);
// 	$rtf->getDocument();

class RTF {

	// {\colortbl;\red 0\green 0\blue 0;\red 255\green 0\ blue0;\red0 ...}
	var $colour_table = array();
	var $colour_rgb;
	// {\fonttbl{\f0}{\f1}{f...}}
	var $font_table = array();
	var $font_face;
	var $font_size;
	// {\info {\title <title>} {\author <author>} {\operator <operator>}}
	var $info_table = array();
	var $page_width;
	var $page_height;
	var $page_size;
	var $page_orientation;
	var $rtf_version;
	var $tab_width;
	
	var $document;
	var $buffer;
	
	var $inch;
	var $mm;
	
	var $fonts_array;
	
	function RTF($paper_orientation=1, $paper_size=5, $font_face=0, $font_size=24, $rtf_version=1, $tab_width=360) {
		
		//$font_face = 0;
		//$font_size = 24;
		//$rtf_version = 1;
		//$tab_width = 360;
		//$paper_size = 5;
		//$paper_orientation = 1;

		// Measurements
		// 1 inch = 1440 twips
		// 1 cm = 567 twips
		// 1 mm = 56.7 twips
		$this->inch = 1440;
		$this->mm = 56.7;

		$this->setDefaultFontFace($font_face);
		$this->setDefaultFontSize($font_size);
		$this->setPaperSize($paper_size);
		$this->setPaperOrientation($paper_orientation);
		$this->rtf_version = $rtf_version;
		$this->tab_width = $tab_width;
		
		// Fonts
		$this->fonts_array = array();
		// Array structure - array(
		//	"name"		=>	Name given to the font,
		//	"family"	=>	[nil, roman, swiss, modern, script, decor, tech, bidi],
		//	"charset"	=>	0
		// );
		
		$this->fonts_array[] = array(
			"name"		=>	"Arial",
			"family"	=>	"swiss",
			"charset"	=>	0
		);
		
		$this->fonts_array[] = array(
			"name"		=>	"Times New Roman",
			"family"	=>	"roman",
			"charset"	=>	0
		);
		
		$this->fonts_array[] = array(
			"name"		=>	"Verdana",
			"family"	=>	"swiss",
			"charset"	=>	0
		);
		
		$this->fonts_array[] = array(
			"name"		=>	"Symbol",
			"family"	=>	"roman",
			"charset"	=>	2
		);

	}
	
	function setDefaultFontFace($face) {
		$this->font_face = $face; // $font is integer
	}
	
	function setDefaultFontSize($size) {
		$this->font_size = $size;
	}
	
	function setTitle($title="") {
		$this->info_table["title"] = $title;
	}
	
	function setAuthor($author="") {
		$this->info_table["author"] = $author;
	}
	
	function setOperator($operator="") {
		$this->info_table["operator"] = $operator;
	}
	
	function setPaperSize($size=0) {
		
		// 1 => Letter (8.5 x 11 inch)
		// 2 => Legal (8.5 x 14 inch)
		// 3 => Executive (7.25 x 10.5 inch)
		// 4 => A3 (297 x 420 mm)
		// 5 => A4 (210 x 297 mm)
		// 6 => A5 (148 x 210 mm)
		// Orientation considered as Portrait
		
		switch($size) {
			case 1:
				$this->page_width = floor(8.5*$this->inch);
				$this->page_height = floor(11*$this->inch);
				$this->page_size = 1;
				break;	
			case 2:
				$this->page_width = floor(8.5*$this->inch);
				$this->page_height = floor(14*$this->inch);
				$this->page_size = 5;
				break;	
			case 3:
				$this->page_width = floor(7.25*$this->inch);
				$this->page_height = floor(10.5*$this->inch);
				$this->page_size = 7;
				break;	
			case 4:
				$this->page_width = floor(297*$this->mm);
				$this->page_height = floor(420*$this->mm);
				$this->page_size = 8;
				break;	
			case 5:
			default:
				$this->page_width = floor(210*$this->mm);
				$this->page_height = floor(297*$this->mm);
				$this->page_size = 9;
				break;	
			case 6:
				$this->page_width = floor(148*$this->mm);
				$this->page_height = floor(210*$this->mm);
				$this->page_size = 10;
				break;	
		}
	}
	
	function setPaperOrientation($orientation=0) {
		// 1 => Portrait
		// 2 => Landscape
		
		switch($orientation) {
			case 1:
			default:
				$this->page_orientation = 1;
				break;
			case 2:
				$this->page_orientation = 2;
				break;	
		}
	}
	
	function addColour($hexcode) {
		// Get the RGB values
		$this->_hex2rgb($hexcode);
		
		// Register in the colour table array
		$this->colour_table[] = array(
			"red"	=>	$this->colour_rgb["red"],
			"green"	=>	$this->colour_rgb["green"],
			"blue"	=>	$this->colour_rgb["blue"]
		);
	}
	
	// Add a text string to the document buffer
	function addText($text, $isUtf8=false) {
		$text = str_replace("\n", "", $text);
		$text = str_replace("\t", "", $text);
		$text = str_replace("\r", "", $text);
		
		$this->document .= $this->_parseText($text, $isUtf8);
	}
	
	// Ouput the RTF file
	function getDocument($filename = 'doc.rtf') {
		$this->buffer .= "{";
		// Header
		$this->buffer .= $this->_getHeader();
		// Font table
		$this->buffer .= $this->_getFontTable();
		// Colour table
		$this->buffer .= $this->_getColourTable();
		// File Information
		$this->buffer .= $this->_getInformation();
		// Default font values
		$this->buffer .= $this->_getDefaultFont();
		// Page display settings
		$this->buffer .= $this->_getPageSettings();
		$this->buffer .= $this->document;
		$this->buffer .= "}";
		
		header("Content-Type: text/enriched\n");
		header("Content-Disposition: attachment; filename=$filename");
		echo $this->buffer;
	}
	
################
# PRIVATE
################

	// Convert HEX to RGB (#FFFFFF => r255 g255 b255)
	function _hex2rgb($hexcode) {
		$hexcode = str_replace("#", "", $hexcode); 
		$rgb = array();
		$rgb["red"] = hexdec(substr($hexcode, 0, 2));
		$rgb["green"] = hexdec(substr($hexcode, 2, 2));
		$rgb["blue"] = hexdec(substr($hexcode, 4, 2));
		
		$this->colour_rgb = $rgb;
	}
	
	// Convert newlines into \par
	function _nl2par($text) {
		$text = str_replace("\n", "\\par ", $text);
		
		return $text;
	}
		
	// Header
	function _getHeader() {
		$header_buffer = "\\rtf{$this->rtf_version}\\ansi\\deff0\\deftab{$this->tab_width}\n\n";
		
		return $header_buffer;
	}
	
	// Font table
	function _getFontTable() {		
		$font_buffer = "{\\fonttbl\n";
		foreach($this->fonts_array AS $fnum => $farray) {
			$font_buffer .= "{\\f{$fnum}\\f{$farray['family']}\\fcharset{$farray['charset']} {$farray['name']}}\n";
		}
		$font_buffer .= "}\n\n";
		
		return $font_buffer;
	}
	
	// Colour table
	function _getColourTable() {
		$colour_buffer = "";
		if(sizeof($this->colour_table) > 0) {
			$colour_buffer = "{\\colortbl;\n";
			foreach($this->colour_table AS $cnum => $carray) {
				$colour_buffer .= "\\red{$carray['red']}\\green{$carray['green']}\\blue{$carray['blue']};\n";	
			}
			$colour_buffer .= "}\n\n";
		}
		
		return $colour_buffer;
	}
	
	// Information
	function _getInformation() {
		$info_buffer = "";
		if(sizeof($this->info_table) > 0) {
			$info_buffer = "{\\info\n";
			foreach($this->info_table AS $name => $value) {
				$info_buffer .= "{\\{$name} {$value}}";
			}
			$info_buffer .= "}\n\n";
		}
		
		return $info_buffer;
	}
	
	// Default font settings
	function _getDefaultFont() {
		$font_buffer = "\\f{$this->font_face}\\fs{$this->font_size}\n";
		
		return $font_buffer;
	}
	
	// Page display settings
	function _getPageSettings() {
		if($this->page_orientation == 1)
			$page_buffer = "\\paperw{$this->page_width}\\paperh{$this->page_height}\n";
		else
			$page_buffer = "\\paperw{$this->page_height}\\paperh{$this->page_width}\\landscape\n";
			
		$page_buffer .= "\\pgncont\\pgndec\\pgnstarts1\\pgnrestart\n";
		
		return $page_buffer;
	}
	
	// Convert special characters to ASCII
	function _specialCharacters($text) {
		$text_buffer = '';
		$cnt = strlen($text);
		for ($i=0;$i<$cnt;$i++){
			$character = $text[$i];
			$ord = ord($character);
			if (($ord >= 0x20 && $ord < 0x80) || $ord == 0x09 || $ord == 0x0A)
				$text_buffer .= $character;
			elseif($ord >= 0x00 && $ord < 0x20)
				$text_buffer .= "\\'".dechex($ord);
			elseif ($ord >= 0x80 && $ord < 0xFF)
				$text_buffer .= "\\'".dechex($ord);
			switch(ord($character)) {
				case 0x5C:
				case 0x7B:
				case 0x7D:
					$text_buffer .= "\\".$character;
					break;
			}
		}
		return $text_buffer;
	}
	
	// Convert special characters to ASCII
	function _specialCharactersUni($text) {
		$text_buffer = '';
		$codepoints = $this->_utf8CodePoints($text);
		#print_r($codepoints);
		$cnt = count($codepoints);
		for ($i=0;$i<$cnt;$i++){
			$ord = $codepoints[$i];
			if (($ord >= 0x20 && $ord < 0x80) || $ord == 0x09 || $ord == 0x0A)
				$text_buffer .= chr($ord);
			elseif($ord >= 0x00 && $ord < 0x20)
				$text_buffer .= "\\'".dechex($ord);
			elseif ($ord >= 0x80 && $ord < 0xFF)
				$text_buffer .= "\\'".dechex($ord);
			else
				$text_buffer .= "\\u".$ord.'_';
			switch($ord) {
				case 0x5C:
				case 0x7B:
				case 0x7D:
					$text_buffer .= "\\".chr($ord);
					break;
			}
		}
		return $text_buffer;
	}
	
	function _utf8CodePoints($utf8string)	{
		$result=array();
		
		$utf8stringlen=strlen($utf8string);
		$utf8step=0;
		$utf8len=0;
		$utf8val=0;
		
		for($idx=0;$idx<$utf8stringlen;$idx++){
			$asc=ord($utf8string[$idx]);
			if($asc<128){
				$utf8step=0;
				$utf8len=0;
				$utf8val=0;
				$result[]=$asc;	
			}else{
				if(($asc & 0xC0) === 0x80){
					if($utf8step>0){
						$utf8val=($utf8val * 64) + ($asc & 0x3F);
						$utf8step++;
						if($utf8step>=$utf8len){
							$utfstep=0;
							$result[]=$utf8val;
						}
					}
				}else{
					$utf8step=0;
				}
				if($utf8step===0){
					if(($asc & 0xE0)===0xC0){	// 2 byte
						$utf8step=1;
						$utf8len=2;
						$utf8val=$asc & 0x1F;
					}else if(($asc & 0xF0)===0xE0){	// 3 byte
						$utf8step=1;
						$utf8len=3;
						$utf8val=$asc & 0x0F;
					}else if(($asc & 0xF8)===0xF0){	// 4 byte
						$utf8step=1;
						$utf8len=4;
						$utf8val=$asc & 0x07;
					}
				}				
			}
		}
		return $result;
	}
		
	// Parse the text input to RTF		
	function _parseText($doc_buffer, $isUtf8=false) {
		if ($isUtf8)
			$doc_buffer = $this->_specialCharactersUni($doc_buffer);
		else
			$doc_buffer = $this->_specialCharacters($doc_buffer);
		
		if(preg_match("/<UL>(.*?)<\/UL>/mi", $doc_buffer)) {
			$doc_buffer = str_replace("<UL>", "", $doc_buffer);
			$doc_buffer = str_replace("</UL>", "", $doc_buffer);
			$doc_buffer = preg_replace("/<LI>(.*?)<\/LI>/mi", "\\f3\\'B7\\tab\\f{$this->font_face} \\1\\par", $doc_buffer);
		}
		
		$doc_buffer = preg_replace("/<P>(.*?)<\/P>/mi", "\\1\\par ", $doc_buffer);
		$doc_buffer = preg_replace("/<B>(.*?)<\/B>/mi", "\\b \\1\\b0 ", $doc_buffer);
		$doc_buffer = preg_replace("/<EM>(.*?)<\/EM>/mi", "\\i \\1\\i0 ", $doc_buffer);
		$doc_buffer = preg_replace("/<U>(.*?)<\/U>/mi", "\\ul \\1\\ul0 ", $doc_buffer);
		$doc_buffer = preg_replace("/<STRIKE>(.*?)<\/STRIKE>/mi", "\\strike \\1\\strike0 ", $doc_buffer);
		$doc_buffer = preg_replace("/<SUB>(.*?)<\/SUB>/mi", "{\\sub \\1}", $doc_buffer);
		$doc_buffer = preg_replace("/<SUP>(.*?)<\/SUP>/mi", "{\\super \\1}", $doc_buffer);
		
		//$doc_buffer = preg_replace("/<H1>(.*?)<\/H1>/mi", "\\pard\\qc\\fs40 \\1\\par\\pard\\fs{$this->font_size} ", $doc_buffer);
		//$doc_buffer = preg_replace("/<H2>(.*?)<\/H2>/mi", "\\pard\\qc\\fs32 \\1\\par\\pard\\fs{$this->font_size} ", $doc_buffer);
		
		$doc_buffer = preg_replace("/<H1>(.*?)<\/H1>/mi", "\\fs48\\b \\1\\b0\\fs{$this->font_size}\\par ", $doc_buffer);
		$doc_buffer = preg_replace("/<H2>(.*?)<\/H2>/mi", "\\fs36\\b \\1\\b0\\fs{$this->font_size}\\par ", $doc_buffer);
		$doc_buffer = preg_replace("/<H3>(.*?)<\/H3>/mi", "\\fs27\\b \\1\\b0\\fs{$this->font_size}\\par ", $doc_buffer);
		
		
		$doc_buffer = preg_replace("/<HR(.*?)>/i", "\\brdrb\\brdrs\\brdrw30\\brsp20 \\pard\\par ", $doc_buffer);
		$doc_buffer = str_replace("<BR>", "\\par ", $doc_buffer);
		$doc_buffer = str_replace("<TAB>", "\\tab ", $doc_buffer);
		
		$doc_buffer = $this->_nl2par($doc_buffer);
		
		return $doc_buffer;
	}
}

?>